
local mod = ...



-- Node: Find nearest Water source --------------------------------------------------------------------------------------------------------
mod:registerBehaviorTreeNode({
    Name = "BEHAVIOR_NODE_FIRE_FINDWATER",
    VariableList = {
		AgentData = "BEHAVIOR_TREE_DATA_AGENT",
        WaterPosition = "BEHAVIOR_TREE_DATA_LOCATION",
    },

    Update = function(self, instance)

        local compAgent = nil
        compAgent = self.AgentData.Agent

        local posAgent = nil
        posAgent = compAgent:getOwner():getGlobalPosition()

        local moveSpot = nil
        moveSpot = self.WaterPosition


        -- find the nearest fire
        local compFire = nil
        local objNearestFire = nil
        local compNearestFire = nil
        local distNearestFire = nil
        distNearestFire = 1000

        compAgent:getLevel():getComponentManager("COMP_FIRE"):getAllEnabledComponent():forEach(function(compFire)

            if (compFire.ActiveCount > 0) and (compFire.WaterCount < compFire.ActiveCount) then

                local objFire = compFire:getOwner()
                local posFire = objFire:getGlobalPosition()
                local dist = math.sqrt( (posAgent[1] - posFire[1])^2 + (posAgent[3] - posFire[3])^2 )

                if dist < distNearestFire then
                    objNearestFire = objFire
                    compNearestFire = compFire
                    distNearestFire = dist
                end

            end

        end)


        -- Find the nearest water source that belongs to that fire
        local objNearestWater = nil
        local distNearestWater = nil

        if objNearestFire then

            if compNearestFire.WaterList then
                if #compNearestFire.WaterList > 0 then

                    distNearestWater = 1000
                    for i = 1, #compNearestFire.WaterList do

                        local objWater = compAgent:getLevel():find(compNearestFire.WaterList[i])
                        if objWater then

                            local posWater = objWater:getGlobalPosition()
                            local dist = math.sqrt( (posAgent[1] - posWater[1])^2 + (posAgent[3] - posWater[3])^2 )

                            if dist < distNearestWater then
                                objNearestWater = objWater
                                distNearestWater = dist
                            end

                        end

                    end
                end
            end
        end


        -- Set destination, if it was found
        if objNearestWater then

            -- Set and record destination
            moveSpot:setDestination(objNearestWater)

            return BEHAVIOR_TREE_NODE_RESULT.TRUE

        else

            return BEHAVIOR_TREE_NODE_RESULT.FALSE

        end

    end -- update

})



-- Node: Find nearest active Fire --------------------------------------------------------------------------------------------------------
mod:registerBehaviorTreeNode({
    Name = "BEHAVIOR_NODE_FIRE_FINDFIRE",
    VariableList = {
		AgentData = "BEHAVIOR_TREE_DATA_AGENT",
        FirePosition = "BEHAVIOR_TREE_DATA_LOCATION",
    },

    Update = function(self, instance)

        local compAgent = nil
        compAgent = self.AgentData.Agent

        local posAgent = nil
        posAgent = compAgent:getOwner():getGlobalPosition()

        local moveSpot = nil
        moveSpot = self.FirePosition

        -- find the nearest fire
        local compFire = nil
        local objNearestFire = nil
        local distNearestFire = nil
        distNearestFire = 1000

        compAgent:getLevel():getComponentManager("COMP_FIRE"):getAllEnabledComponent():forEach(function(compFire)

            if (compFire.ActiveCount > 0) and (compFire.WaterCount < compFire.ActiveCount) then

                local objFire = compFire:getOwner()
                local posFire = objFire:getGlobalPosition()
                local dist = math.sqrt( (posAgent[1] - posFire[1])^2 + (posAgent[3] - posFire[3])^2 )

                if dist < distNearestFire then
                    objNearestFire = objFire
                    distNearestFire = dist
                end

            end

        end)


        -- Set destination, if it was found
        if objNearestFire then

            -- Set and record destination
            moveSpot:setDestination(objNearestFire)

            return BEHAVIOR_TREE_NODE_RESULT.TRUE

        else

            return BEHAVIOR_TREE_NODE_RESULT.FALSE

        end

    end -- update

})



-- Node: Add Water to Fire --------------------------------------------------------------------------------------------------------
mod:registerBehaviorTreeNode({
    Name = "BEHAVIOR_NODE_FIRE_ADDWATER",
    VariableList = {
		AgentData = "BEHAVIOR_TREE_DATA_AGENT",
    },

    Update = function(self, instance)

        local compAgent = nil
        compAgent = self.AgentData.Agent

        local posAgent = nil
        posAgent = compAgent:getOwner():getGlobalPosition()

        -- find a nearby fire
        local compFire = nil
        local compNearestFire = nil
        local distNearestFire = nil
        distNearestFire = 1000

        compAgent:getLevel():getComponentManager("COMP_FIRE"):getAllEnabledComponent():forEach(function(compFire)

            if (compFire.ActiveCount > 0) and (compFire.WaterCount < compFire.ActiveCount) then

                local objFire = compFire:getOwner()
                local posFire = objFire:getGlobalPosition()
                local dist = math.sqrt( (posAgent[1] - posFire[1])^2 + (posAgent[3] - posFire[3])^2 )

                if dist < distNearestFire and dist < 2 then
                    compNearestFire = compFire
                    distNearestFire = dist
                end

            end

        end)


        -- Add water, if fire was found
        if compNearestFire then
            compNearestFire.WaterCount = compNearestFire.WaterCount + 1
        end

        -- Return true regardless of the result, so that the villager will continue fighting fires
        return BEHAVIOR_TREE_NODE_RESULT.TRUE

    end -- update

})



---------------------------------------------------------------------------------------------------
-- Behavior Tree ----------------------------------------------------------------------------------
---------------------------------------------------------------------------------------------------
mod:registerBehaviorTree({
	Id = "BEHAVIOR_FIRE_WATER",
	VariableList = {
		{
			Name = "AgentData",
			DataType = "BEHAVIOR_TREE_DATA_AGENT",
			IsPublic = true,
			InitialValue = {}
		},
        {
            Name = "TimeToWait",
            DataType = "BEHAVIOR_TREE_DATA_WAIT",
            IsPublic = false,
            InitialValue = {
                TimeToWait = 1,
            },
        },
        {
			Name = "WaterPosition",
			DataType = "BEHAVIOR_TREE_DATA_LOCATION",
			IsPublic = false,
			InitialValue = {
				CanNavigateOnGround = true,
				CanNavigateOnWater = false,
				IsSetOrientationOnDestination = true
			}
		},
        {
			Name = "FirePosition",
			DataType = "BEHAVIOR_TREE_DATA_LOCATION",
			IsPublic = false,
			InitialValue = {
				CanNavigateOnGround = true,
				CanNavigateOnWater = false,
				IsSetOrientationOnDestination = false
			}
		},
		{
			Name = "WorkLoop",
			DataType = "BEHAVIOR_TREE_DATA_LOOP",
			IsPublic = true,
			InitialValue = {
				LoopCount = 20,
				Duration = 0,
				IsInfinite = false,
				IsDuration = false,
			}
		},
	},
	Root = {
        Name = "WorkLoopRepeater",
        Type = "REPEAT",
        RepeatData = "WorkLoop",
        Child = {
            Name = "WorkLoopSequencer",
            Type = "SEQUENCER",
            Children = {
                {
                    Name = "SetWaterDestination",
                    Type = "BEHAVIOR_NODE_FIRE_FINDWATER",
                    AgentData = "AgentData",
                    WaterPosition = "WaterPosition",
                },
                {
                    Name = "GoToWater",
                    Type = "GO_TO",
                    AgentData = "AgentData",
                    Destination = "WaterPosition",
                    BuildingPathType = "",
                    AnimationData = "",
                    AnimationSpeedMultiplier = ""
                },
                {
                    Name = "WaitAtWater",
                    Type = "WAIT",
                    AgentData = "AgentData",
                    TimeToWait = "TimeToWait",
                    OptionalUseWorkstationAnim = "",
                },
                {
                    Name = "SetFireDestination",
                    Type = "BEHAVIOR_NODE_FIRE_FINDFIRE",
                    AgentData = "AgentData",
                    FirePosition = "FirePosition",
                },
                {
                    Name = "GoToFire",
                    Type = "GO_TO",
                    AgentData = "AgentData",
                    Destination = "FirePosition",
                    BuildingPathType = "",
                    AnimationData = "",
                    AnimationSpeedMultiplier = ""
                },
                {
                    Name = "WaitAtFire",
                    Type = "WAIT",
                    AgentData = "AgentData",
                    TimeToWait = "TimeToWait",
                    OptionalUseWorkstationAnim = "",
                },
                {
                    Name = "AddWaterToFire",
                    Type = "BEHAVIOR_NODE_FIRE_ADDWATER",
                    AgentData = "AgentData",
                },
            }
        }
	}
})